import numpy.testing as npt
import numpy as np
from pyHalo.Halos.HaloModels.TNFW import TNFWFieldHalo, TNFWSubhalo
from pyHalo.Halos.HaloModels.core_collapsed_halo import CoreCollapsedHalo, CoreCollapsedHaloBH
from pyHalo.truncation_models import ConstantTruncationArcsec
from pyHalo.concentration_models import ConcentrationConstant
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.single_realization import Realization
from pyHalo.realization_extensions import RealizationExtensions
import pytest


class TestCoreCollapsedHalo(object):

    def setup_method(self):

        self.lens_cosmo = LensCosmo(0.5, 2.5)
        self.truncation_class_field_halo = ConstantTruncationArcsec(self.lens_cosmo, 1000.0)

    def test_field_halo_highc(self):
        mass = 10 ** 8
        halo_concentration = 25.0
        x = 0.0
        y = 0.0
        r3d = None
        z = 0.5
        sub_flag = False
        lens_cosmo = LensCosmo(z, 2.0)
        unique_tag = 1.0
        tau = 2000
        tnfw_halo = TNFWFieldHalo.simple_setup(mass, x, y, z, tau, lens_cosmo,
                                               concentration_model='DIEMERJOYCE19')
        tnfw_halo._c = halo_concentration
        _, rs, r200 = tnfw_halo.nfw_params
        rt_kpc = tnfw_halo.profile_args[1]
        gamma_match = -2.0
        r_match_kpc = tnfw_halo.log_derivative_inverse(gamma_match, rs, rt_kpc)
        mass_R = tnfw_halo.mass_3d(r_match_kpc)
        Rs_inner_kpc = r_match_kpc
        gamma_inner = 2.5
        gamma_outer = 6.0

        truncation_class = None
        concentration_class = ConcentrationConstant(None, tnfw_halo.c)

        scale_match_m = 1.6
        m_target_R = scale_match_m * mass_R
        args = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'rt_kpc': rt_kpc,
                'm_target_r200': mass, 'm_target_R': m_target_R, 'Rs_inner_kpc': Rs_inner_kpc,
                'r_match_kpc': r_match_kpc}
        halo = CoreCollapsedHalo(mass, x, y, r3d, z,
                                 sub_flag, lens_cosmo, args, truncation_class, concentration_class, unique_tag)

        r = np.logspace(-3.5, np.log10(tnfw_halo.c), 2000) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        total_density = density_inner + density_outer
        density_nfw = tnfw_halo.density_profile_3d(r)
        mass_cc_r200 = np.trapz(4 * np.pi * r ** 2 * total_density, r)
        mass_nfw_r200 = np.trapz(4 * np.pi * r ** 2 * density_nfw, r)
        npt.assert_almost_equal(mass_nfw_r200 / mass, 1,2)
        npt.assert_almost_equal(mass_cc_r200 / mass, 1, 2)

        r = np.logspace(-4, np.log10(r_match_kpc/rs), 2500) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        total_density = density_inner + density_outer
        density_nfw = tnfw_halo.density_profile_3d(r)
        mass_cc_rmatch = np.trapz(4 * np.pi * r ** 2 * total_density, r)
        mass_nfw_rmatch = scale_match_m * np.trapz(4 * np.pi * r ** 2 * density_nfw, r)
        npt.assert_almost_equal(mass_cc_rmatch / mass_nfw_rmatch, 1, 2)

    def test_field_halo_lowc(self):
        mass = 10 ** 8
        halo_concentration = 3.0
        x = 0.0
        y = 0.0
        r3d = None
        z = 0.5
        sub_flag = False
        lens_cosmo = LensCosmo(z, 2.0)
        unique_tag = 1.0
        tau = 2000
        tnfw_halo = TNFWFieldHalo.simple_setup(mass, x, y, z, tau, lens_cosmo,
                                               concentration_model='DIEMERJOYCE19')
        tnfw_halo._c = halo_concentration
        _, rs, r200 = tnfw_halo.nfw_params
        rt_kpc = tnfw_halo.profile_args[1]
        gamma_match = -2.0
        r_match_kpc = tnfw_halo.log_derivative_inverse(gamma_match, rs, rt_kpc)
        mass_R = tnfw_halo.mass_3d(r_match_kpc)
        Rs_inner_kpc = r_match_kpc
        gamma_inner = 2.5
        gamma_outer = 6.0

        truncation_class = None
        concentration_class = ConcentrationConstant(None, tnfw_halo.c)

        scale_match_m = 1.6
        m_target_R = scale_match_m * mass_R
        args = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'rt_kpc': rt_kpc,
                'm_target_r200': mass, 'm_target_R': m_target_R, 'Rs_inner_kpc': Rs_inner_kpc,
                'r_match_kpc': r_match_kpc}
        halo = CoreCollapsedHalo(mass, x, y, r3d, z,
                                 sub_flag, lens_cosmo, args, truncation_class, concentration_class, unique_tag)

        r = np.logspace(-3.5, np.log10(tnfw_halo.c), 2000) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        total_density = density_inner + density_outer
        density_nfw = tnfw_halo.density_profile_3d(r)
        mass_cc_r200 = np.trapz(4 * np.pi * r ** 2 * total_density, r)
        mass_nfw_r200 = np.trapz(4 * np.pi * r ** 2 * density_nfw, r)
        npt.assert_almost_equal(mass_nfw_r200 / mass, 1,2)
        npt.assert_almost_equal(mass_cc_r200 / mass, 1, 2)

        r = np.logspace(-4, np.log10(r_match_kpc/rs), 2500) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        total_density = density_inner + density_outer
        density_nfw = tnfw_halo.density_profile_3d(r)
        mass_cc_rmatch = np.trapz(4 * np.pi * r ** 2 * total_density, r)
        mass_nfw_rmatch = scale_match_m * np.trapz(4 * np.pi * r ** 2 * density_nfw, r)
        npt.assert_almost_equal(mass_cc_rmatch / mass_nfw_rmatch, 1, 2)

    def test_subhalo_low_fbound(self):
        mass = 10 ** 8
        x = 0.0
        y = 0.0
        r3d = None
        z = 0.5
        sub_flag = False
        lens_cosmo = LensCosmo(z, 2.0)
        unique_tag = 1.0
        f_bound = 0.001
        z_infall = 2.0
        tnfw_subhalo = TNFWSubhalo.simple_setup(mass, f_bound, z_infall, x, y, z, lens_cosmo)
        _, rs, r200 = tnfw_subhalo.nfw_params
        rt_kpc = tnfw_subhalo.profile_args[1]
        r_match_kpc = tnfw_subhalo.log_derivative_inverse(-2.0, rs, rt_kpc)
        mass_R = tnfw_subhalo.mass_3d(r_match_kpc)
        m_target_r200 = tnfw_subhalo.mass_3d(r200)
        Rs_inner_kpc = r_match_kpc
        gamma_inner = 2.4
        gamma_outer = 6.0
        scale_match_mass = 2.0
        args = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'rt_kpc': rt_kpc,
                'm_target_r200': m_target_r200, 'm_target_R': scale_match_mass * mass_R, 'r_match_kpc': r_match_kpc,
                'Rs_inner_kpc': Rs_inner_kpc}

        truncation_class = None
        concentration_class = ConcentrationConstant(None, tnfw_subhalo.c)

        halo = CoreCollapsedHalo(mass, x, y, r3d, z,
                                 sub_flag, lens_cosmo, args, truncation_class, concentration_class, unique_tag)

        r = np.logspace(-4, np.log10(tnfw_subhalo.c), 100) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        total_density = density_inner + density_outer
        mass_numerical = np.trapz(4*np.pi*r**2*total_density, r)
        npt.assert_array_less(abs(-1+m_target_r200 / mass_numerical), 0.03)

        r = np.logspace(-4, np.log10(r_match_kpc/rs), 2000) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        total_density = density_inner + density_outer
        mass_numerical = np.trapz(4 * np.pi * r ** 2 * total_density, r)
        npt.assert_array_less(abs(-1+scale_match_mass * mass_R / mass_numerical), 0.03)

    def test_subhalo_high_fbound(self):

        mass = 10 ** 8
        x = 0.0
        y = 0.0
        r3d = None
        z = 0.5
        sub_flag = False
        lens_cosmo = LensCosmo(z, 2.0)
        unique_tag = 1.0
        f_bound = 0.6
        z_infall = 2.0
        tnfw_subhalo = TNFWSubhalo.simple_setup(mass, f_bound, z_infall, x, y, z, lens_cosmo)
        _, rs, r200 = tnfw_subhalo.nfw_params
        rt_kpc = tnfw_subhalo.profile_args[1]
        r_match_kpc = tnfw_subhalo.log_derivative_inverse(-2.0, rs, rt_kpc)
        mass_R = tnfw_subhalo.mass_3d(r_match_kpc)
        m_target_r200 = tnfw_subhalo.mass_3d(r200)
        Rs_inner_kpc = r_match_kpc
        gamma_inner = 2.4
        gamma_outer = 6.0
        scale_match_mass = 2.
        m_target_R = scale_match_mass * mass_R
        args = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'rt_kpc': rt_kpc,
                'm_target_r200': m_target_r200, 'm_target_R': m_target_R, 'r_match_kpc': r_match_kpc,
                'Rs_inner_kpc': Rs_inner_kpc}

        truncation_class = None
        concentration_class = ConcentrationConstant(None, tnfw_subhalo.c)

        halo = CoreCollapsedHalo(mass, x, y, r3d, z,
                                 sub_flag, lens_cosmo, args, truncation_class, concentration_class, unique_tag)

        r = np.logspace(-3.5, np.log10(tnfw_subhalo.c), 1000) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        m_num = np.trapz(4 * np.pi * r ** 2 * (density_inner + density_outer), r)
        npt.assert_array_less(abs(-1+m_num / m_target_r200), 0.07)

        r = np.logspace(-3.5, np.log10(r_match_kpc / rs), 100) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        m_num = np.trapz(4 * np.pi * r ** 2 * (density_inner + density_outer), r)
        npt.assert_array_less(abs(-1+m_num / m_target_R), 0.03)

    def test_subhalo_low_fbound(self):

        mass = 10 ** 8
        x = 0.0
        y = 0.0
        r3d = None
        z = 0.5
        sub_flag = False
        lens_cosmo = LensCosmo(z, 2.0)
        unique_tag = 1.0
        f_bound = 0.005
        z_infall = 2.0
        tnfw_subhalo = TNFWSubhalo.simple_setup(mass, f_bound, z_infall, x, y, z, lens_cosmo)
        _, rs, r200 = tnfw_subhalo.nfw_params
        rt_kpc = tnfw_subhalo.profile_args[1]
        r_match_kpc = tnfw_subhalo.log_derivative_inverse(-2.0, rs, rt_kpc)
        mass_R = tnfw_subhalo.mass_3d(r_match_kpc)
        m_target_r200 = tnfw_subhalo.mass_3d(r200)
        Rs_inner_kpc = r_match_kpc
        gamma_inner = 2.4
        gamma_outer = 6.0
        scale_match_mass = 2.
        m_target_R = scale_match_mass * mass_R
        args = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'rt_kpc': rt_kpc,
                'm_target_r200': m_target_r200, 'm_target_R': m_target_R, 'r_match_kpc': r_match_kpc,
                'Rs_inner_kpc': Rs_inner_kpc}

        truncation_class = None
        concentration_class = ConcentrationConstant(None, tnfw_subhalo.c)

        halo = CoreCollapsedHalo(mass, x, y, r3d, z,
                                 sub_flag, lens_cosmo, args, truncation_class, concentration_class, unique_tag)

        r = np.logspace(-3.5, np.log10(tnfw_subhalo.c), 1000) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        m_num = np.trapz(4 * np.pi * r ** 2 * (density_inner + density_outer), r)
        npt.assert_array_less(abs(-1+m_num / m_target_r200), 0.03)

        r = np.logspace(-3.5, np.log10(r_match_kpc / rs), 1000) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        m_num = np.trapz(4 * np.pi * r ** 2 * (density_inner + density_outer), r)
        npt.assert_array_less(abs(-1+m_num / m_target_R), 0.04)

    def test_BH(self):

        mass = 10 ** 8
        x = 0.0
        y = 0.0
        r3d = None
        z = 0.5
        sub_flag = False
        lens_cosmo = LensCosmo(z, 2.0)
        unique_tag = 1.0
        f_bound = 0.005
        z_infall = 2.0
        tnfw_subhalo = TNFWSubhalo.simple_setup(mass, f_bound, z_infall, x, y, z, lens_cosmo)
        _, rs, r200 = tnfw_subhalo.nfw_params
        rt_kpc = tnfw_subhalo.profile_args[1]
        r_match_kpc = tnfw_subhalo.log_derivative_inverse(-2.0, rs, rt_kpc)
        mass_R = tnfw_subhalo.mass_3d(r_match_kpc)
        m_target_r200 = tnfw_subhalo.mass_3d(r200)
        Rs_inner_kpc = r_match_kpc
        gamma_inner = 2.4
        gamma_outer = 6.0
        scale_match_mass = 2.
        m_target_R = scale_match_mass * mass_R
        args = {'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer, 'rt_kpc': rt_kpc,
                'm_target_r200': m_target_r200, 'm_target_R': m_target_R, 'r_match_kpc': r_match_kpc,
                'Rs_inner_kpc': Rs_inner_kpc}

        truncation_class = None
        concentration_class = ConcentrationConstant(None, tnfw_subhalo.c)

        halo_profile = CoreCollapsedHalo(mass, x, y, r3d, z,
                                 sub_flag, lens_cosmo, args, truncation_class, concentration_class, unique_tag)
        halo_bh = CoreCollapsedHaloBH(mass, x, y, r3d, z,
                                 sub_flag, lens_cosmo, args, truncation_class, concentration_class, unique_tag)

        kwargs_profile = halo_profile.lenstronomy_params_split[0]
        kwargs_bh = halo_bh.lenstronomy_params[0][0]
        kwargs_envelope_profile = kwargs_profile[1]
        npt.assert_almost_equal(kwargs_envelope_profile['Rs'], kwargs_bh['Rs_outer'])
        npt.assert_almost_equal(kwargs_envelope_profile['alpha_Rs'], kwargs_bh['alpha_Rs_outer'])
        npt.assert_almost_equal(kwargs_envelope_profile['r_trunc'], kwargs_bh['r_trunc'])

        # calculate the mass of the central profile
        theta_E_target = lens_cosmo.point_mass_factor_z(halo_bh.z) * np.sqrt(m_target_R)
        npt.assert_almost_equal(kwargs_bh['theta_E_inner'], theta_E_target)

    def test_through_realization_extensions(self):

        mass = 10 ** 9
        x = 0.0
        y = 0.0
        z = 0.5
        lens_cosmo = LensCosmo(z, 2.0)
        f_bound = 0.6
        z_infall = 2.0
        kwargs_halo_model = None
        msheet_correction = None
        rendering_classes = None
        rendering_center_x = 0.0
        rendering_center_y = 0.0
        geometry = None
        tnfw_subhalo = TNFWSubhalo.simple_setup(mass, f_bound, z_infall, x, y, z, lens_cosmo)
        _, rs, r200 = tnfw_subhalo.nfw_params
        rt_kpc = tnfw_subhalo.profile_args[1]
        log_slope_match = -2.1
        r_match_kpc = tnfw_subhalo.log_derivative_inverse(log_slope_match, rs, rt_kpc)
        mass_R = tnfw_subhalo.mass_3d(r_match_kpc)
        m_target_r200 = tnfw_subhalo.mass_3d(r200)
        scale_match_mass = 2.5
        gamma_inner = 2.5
        m_target_R = scale_match_mass * mass_R
        halo_realization = Realization.from_halos([tnfw_subhalo], lens_cosmo, kwargs_halo_model,
                                                  msheet_correction, rendering_classes, rendering_center_x,
                                                  rendering_center_y, geometry)
        ext = RealizationExtensions(halo_realization)
        kwargs_halo = {'gamma_inner': gamma_inner, 'scale_match_r': scale_match_mass, 'log_slope_match': log_slope_match}
        realization_sidm = ext.add_core_collapsed_halos([0], halo_profile='CC_COMPOSITE', **kwargs_halo)
        halo = realization_sidm.halos[0]

        r = np.logspace(-3.5, np.log10(tnfw_subhalo.c), 1000) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        m_num = np.trapz(4 * np.pi * r ** 2 * (density_inner + density_outer), r)
        npt.assert_array_less(abs(-1 + m_num / m_target_r200), 0.03)

        r = np.logspace(-3.5, np.log10(r_match_kpc / rs), 1000) * rs
        density_inner, density_outer = halo.component_density_profile_3d(r)
        m_num = np.trapz(4 * np.pi * r ** 2 * (density_inner + density_outer), r)
        npt.assert_array_less(abs(-1 + m_num / m_target_R), 0.04)

    def test_through_realization_extensions_BH(self):
        mass = 10 ** 9
        x = 0.0
        y = 0.0
        z = 0.5
        lens_cosmo = LensCosmo(z, 2.0)
        f_bound = 0.6
        z_infall = 2.0
        kwargs_halo_model = None
        msheet_correction = None
        rendering_classes = None
        rendering_center_x = 0.0
        rendering_center_y = 0.0
        geometry = None
        tnfw_subhalo = TNFWSubhalo.simple_setup(mass, f_bound, z_infall, x, y, z, lens_cosmo)
        _, rs, r200 = tnfw_subhalo.nfw_params
        log_slope_match = -2.1
        scale_match_mass = 2.5
        gamma_inner = -1
        halo_realization = Realization.from_halos([tnfw_subhalo], lens_cosmo, kwargs_halo_model,
                                                  msheet_correction, rendering_classes, rendering_center_x,
                                                  rendering_center_y, geometry)
        ext = RealizationExtensions(halo_realization)
        kwargs_halo = {'gamma_inner': gamma_inner,
                       'scale_match_r': scale_match_mass,
                       'log_slope_match': log_slope_match}
        realization_sidm = ext.add_core_collapsed_halos([0],
                                                        halo_profile='CC_COMPOSITE',
                                                        **kwargs_halo)
        halo = realization_sidm.halos[0]
        lenstronomy_ID = halo.lenstronomy_ID
        npt.assert_string_equal(lenstronomy_ID[0], 'CORE_COLLAPSED_HALO_BH')

if __name__ == '__main__':
    pytest.main()



import numpy.testing as npt
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Halos.HaloModels.TNFW import TNFWSubhalo
import numpy as np
from pyHalo.single_realization import Realization
from pyHalo.realization_extensions import RealizationExtensions
import pytest

class TestSubhaloTransform(object):

    def setup_method(self):

        self.lens_cosmo = LensCosmo(0.5, 2.5)

    def test_transform_heavy_stripping(self):

        x = 0.5
        y = 1.0
        z = 0.45
        z_infall = 4.0
        mass = 1*10 ** 7
        f_bound = 0.001
        tnfw_subhalo = TNFWSubhalo.simple_setup(mass,
                                                f_bound,
                                                z_infall,
                                                x,
                                                y,
                                                z,
                                                self.lens_cosmo)
        realization = Realization.from_halos([tnfw_subhalo],
                                             self.lens_cosmo,
                                             kwargs_halo_model=None,
                                             msheet_correction=None,
                                             rendering_classes=None)
        ext = RealizationExtensions(realization)
        indexes = [0]
        x_core_halo = 0.05
        sidm_subhalo = ext.add_core_collapsed_halos(indexes,
                                           t_c=None,
                                           x_core_halo=x_core_halo).halos[0]

        #print(tnfw_subhalo.rescale_norm, tnfw_subhalo.profile_args[1]/tnfw_subhalo.nfw_params[1])
        mass_tnfw_subhalo = tnfw_subhalo.mass_3d('r200')
        mass_sidm_subhalo = sidm_subhalo.mass_3d('r200')
        npt.assert_almost_equal(mass_tnfw_subhalo / mass_sidm_subhalo, 1, 2)
        rhos, rs, r200 = tnfw_subhalo.nfw_params
        r = np.linspace(0.001 * r200, r200, 1000)
        density_nfw = tnfw_subhalo.density_profile_3d_lenstronomy(r)
        density_sidm = sidm_subhalo.density_profile_3d_lenstronomy(r)
        mass_analytic_tnfw = np.trapz(4 * np.pi * r ** 2 * density_nfw, r)
        mass_analytic_sidm = np.trapz(4 * np.pi * r ** 2 * density_sidm, r)
        npt.assert_almost_equal(mass_analytic_tnfw / mass_analytic_sidm, 1, 2)

    def test_transform_light_stripping(self):

        x = 0.5
        y = 1.0
        z = 0.45
        z_infall = 1.0
        mass = 1*10 ** 7.7
        f_bound = 0.9
        tnfw_subhalo = TNFWSubhalo.simple_setup(mass,
                                                f_bound,
                                                z_infall,
                                                x,
                                                y,
                                                z,
                                                self.lens_cosmo)
        realization = Realization.from_halos([tnfw_subhalo],
                                             self.lens_cosmo,
                                             kwargs_halo_model=None,
                                             msheet_correction=None,
                                             rendering_classes=None)
        ext = RealizationExtensions(realization)
        indexes = [0]
        x_core_halo = 0.05
        sidm_subhalo = ext.add_core_collapsed_halos(indexes,
                                           t_c=None,
                                           x_core_halo=x_core_halo).halos[0]

        #print(tnfw_subhalo.rescale_norm, tnfw_subhalo.profile_args[1]/tnfw_subhalo.nfw_params[1])
        mass_tnfw_subhalo = tnfw_subhalo.mass_3d('r200')
        mass_sidm_subhalo = sidm_subhalo.mass_3d('r200')
        npt.assert_almost_equal(mass_tnfw_subhalo / mass_sidm_subhalo, 1, 2)
        rhos, rs, r200 = tnfw_subhalo.nfw_params
        r = np.linspace(0.001 * r200, r200, 1000)
        density_nfw = tnfw_subhalo.density_profile_3d_lenstronomy(r)
        density_sidm = sidm_subhalo.density_profile_3d_lenstronomy(r)
        mass_analytic_tnfw = np.trapz(4 * np.pi * r ** 2 * density_nfw, r)
        mass_analytic_sidm = np.trapz(4 * np.pi * r ** 2 * density_sidm, r)
        npt.assert_almost_equal(mass_analytic_tnfw / mass_analytic_sidm, 1, 2)

if __name__ == '__main__':
    pytest.main()



import numpy as np
from lenstronomy.LensModel.Profiles.tnfw import TNFW
from lenstronomy.LensModel.Profiles.pseudo_double_powerlaw import PseudoDoublePowerlaw
from pyHalo.Halos.halo_base import Halo
from copy import deepcopy

class CoreCollapsedHalo(Halo):
    _pseudo_nfw = False
    """

    """

    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._profile_envelope = TNFW()
        self._profile_center = PseudoDoublePowerlaw()
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        self._penalty_final_R200, self._penalty_final_R = None, None
        mdef = 'CORE_COLLAPSED'
        super(CoreCollapsedHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                                lens_cosmo_instance, args, unique_tag)

    def component_density_profile_3d(self, r, profile_args=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        kwargs_lenstronomy = deepcopy(self.lenstronomy_params_split[0])
        kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
        sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
        sigma_crit_kpc = sigma_crit_mpc * 1e-6
        factor = sigma_crit_kpc / kpc_per_arcsec
        del kwargs_lenstronomy[0]['center_x']
        del kwargs_lenstronomy[0]['center_y']
        del kwargs_lenstronomy[1]['center_x']
        del kwargs_lenstronomy[1]['center_y']
        inner = factor * self._profile_center.density_lens(r / kpc_per_arcsec, **kwargs_lenstronomy[0])
        kwargs_lenstronomy[1]['rho0'] = self._profile_envelope.alpha2rho0(kwargs_lenstronomy[1]['alpha_Rs'],
                                                                          kwargs_lenstronomy[1]['Rs'])
        del kwargs_lenstronomy[1]['alpha_Rs']
        outer = factor * self._profile_envelope.density(r / kpc_per_arcsec, **kwargs_lenstronomy[1])
        return inner, outer

    def density_profile_3d(self, r, kwargs_lenstronomy=None):
        inner, outer = self.component_density_profile_3d(r, kwargs_lenstronomy)
        return inner + outer

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def lenstronomy_params_split(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args_split'):
            (alpha_Rs_center, Rs_angle, Rs_angle_inner, gamma_inner,
             gamma_outer, alpha_Rs_envelope, Rs_angle, r_trunc_arcsec) = self.profile_args

            x, y = np.round(self.x, 4), np.round(self.y, 4)
            Rs_angle = np.round(Rs_angle, 10)
            alpha_Rs_center = np.round(alpha_Rs_center, 10)
            Rs_angle_inner = np.round(Rs_angle_inner, 10)
            kwargs_collpsed_center = {'alpha_Rs': alpha_Rs_center,
                                      'Rs': Rs_angle_inner,
                                      'center_x': x,
                                      'center_y': y,
                                      'gamma_inner': gamma_inner,
                                      'gamma_outer': gamma_outer}

            r_trunc_arcsec = np.round(r_trunc_arcsec, 10)
            alpha_Rs_envelope = np.round(alpha_Rs_envelope, 10)
            kwargs_envelope = {'alpha_Rs': alpha_Rs_envelope,
                               'Rs': Rs_angle,
                               'center_x': x,
                               'center_y': y,
                               'r_trunc': r_trunc_arcsec}
            self._lenstronomy_args_split = [kwargs_collpsed_center, kwargs_envelope]
        return self._lenstronomy_args_split, None

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args'):
            (alpha_Rs_center, Rs_angle, Rs_angle_inner, gamma_inner,
             gamma_outer, alpha_Rs_envelope, Rs_angle, r_trunc_arcsec) = self.profile_args

            x, y = np.round(self.x, 4), np.round(self.y, 4)
            Rs_angle = np.round(Rs_angle, 10)
            alpha_Rs_center = np.round(alpha_Rs_center, 10)
            Rs_angle_inner = np.round(Rs_angle_inner, 10)
            r_trunc_arcsec = np.round(r_trunc_arcsec, 10)
            alpha_Rs_envelope = np.round(alpha_Rs_envelope, 10)
            kwargs = {'center_x': x, 'center_y': y, 'Rs_inner': Rs_angle_inner, 'Rs_outer': Rs_angle,
                      'alpha_Rs_inner': alpha_Rs_center, 'alpha_Rs_outer': alpha_Rs_envelope,
                      'r_trunc': r_trunc_arcsec, 'gamma_inner': gamma_inner, 'gamma_outer': gamma_outer}
            self._lenstronomy_args = [kwargs]
        return self._lenstronomy_args, None

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['CORE_COLLAPSED_HALO']

    @property
    def profile_args(self):
        """

        :return:
        """
        if not hasattr(self, '_profile_args'):
            # solve for lens model parameters that conserve mass inside rs and r200
            gamma_inner = self._args['gamma_inner']
            gamma_outer = self._args['gamma_outer']
            rt_kpc = self._args['rt_kpc']
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
            sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2
            rhos, rs, r200 = self.nfw_params
            Rs_angle = rs / kpc_per_arcsec
            r_trunc_angle = rt_kpc / kpc_per_arcsec
            Rs_angle_inner = self._args['Rs_inner_kpc'] / kpc_per_arcsec

            rho0 = self._profile_envelope.alpha2rho0(1.0, Rs_angle)
            r_match_arcsec = self._args['r_match_kpc'] / kpc_per_arcsec
            r200_arcsec = r200 / kpc_per_arcsec
            a = self._profile_envelope.mass_3d(r_match_arcsec,
                                               Rs_angle,
                                               rho0,
                                               r_trunc_angle/Rs_angle) # mass enclosed NFW inside r_match
            b = self._profile_center.mass_3d_lens(r_match_arcsec,
                                                  Rs_angle_inner,
                                                  1.0,
                                                  gamma_inner,
                                                  gamma_outer)  # mass enclosed NFW inside r_match
            c = self._profile_envelope.mass_3d(r200_arcsec,
                                               Rs_angle,
                                               rho0,
                                               r_trunc_angle / Rs_angle)  # mass enclosed NFW inside r_match
            d = self._profile_center.mass_3d_lens(r200_arcsec,
                                                  Rs_angle_inner,
                                                  1.0,
                                                  gamma_inner,
                                                  gamma_outer)  # mass enclosed NFW inside r_match
            x = np.array([self._args['m_target_R'], self._args['m_target_r200']])
            M_inverse = (1 / (a * d - b * c)) * np.array([[d, -b], [-c, a]])
            y = np.matmul(M_inverse, x) / sigma_crit_arcsec
            alpha_Rs_envelope, alpha_Rs_center = y[0], y[1]
            if alpha_Rs_center < 0 or alpha_Rs_envelope < 0:
                raise Exception('alpha_Rs_envelope and alpha_Rs_center cannot be negative')
            assert np.isfinite(alpha_Rs_envelope)
            assert np.isfinite(alpha_Rs_center)
            self._profile_args = (alpha_Rs_center, Rs_angle, Rs_angle_inner, gamma_inner,
                                  gamma_outer, alpha_Rs_envelope, Rs_angle, r_trunc_angle)

        return self._profile_args

    @property
    def halo_effective_age(self):
        if not hasattr(self, '_halo_effective_age'):
            if self.is_subhalo:
                self._halo_effective_age = self.lens_cosmo.sidm_halo_effective_age(self.z,
                                                                                   self.z_infall,
                                                                                   self._args['lambda_t'])
            else:
                self._halo_effective_age = self.halo_age
        return self._halo_effective_age

    @property
    def sidm_timescale(self):
        """
        Computes the timescale given by Equation 2.2 in https://arxiv.org/pdf/2305.16176.pdf
        :return:
        """
        return self._args['sidm_timescale']

    @property
    def t_over_tc(self):
        """
        Computes the dimensionless timescale for the halo evolution
        :return:
        """
        return self.halo_effective_age / self.sidm_timescale

class CoreCollapsedHaloBH(CoreCollapsedHalo):

    @property
    def lenstronomy_params_split(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args_split'):
            (theta_E_inner, Rs_angle, alpha_Rs_envelope, Rs_angle, r_trunc_arcsec) = self.profile_args

            x, y = np.round(self.x, 4), np.round(self.y, 4)
            Rs_angle = np.round(Rs_angle, 10)
            r_trunc_arcsec = np.round(r_trunc_arcsec, 10)
            alpha_Rs_envelope = np.round(alpha_Rs_envelope, 10)
            theta_E_inner = np.round(theta_E_inner, 10)
            kwargs_envelope = {'center_x': x,
                               'center_y': y,
                               'Rs': Rs_angle,
                        'alpha_Rs': alpha_Rs_envelope,
                        'r_trunc': r_trunc_arcsec}
            kwargs_center = {'theta_E': theta_E_inner, 'center_x': x, 'cetner_y': y}
            self._lenstronomy_args = [kwargs_center, kwargs_envelope]
        return self._lenstronomy_args, None

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args'):
            (theta_E_inner, Rs_angle, alpha_Rs_envelope, Rs_angle, r_trunc_arcsec) = self.profile_args

            x, y = np.round(self.x, 4), np.round(self.y, 4)
            Rs_angle = np.round(Rs_angle, 10)
            r_trunc_arcsec = np.round(r_trunc_arcsec, 10)
            alpha_Rs_envelope = np.round(alpha_Rs_envelope, 10)
            theta_E_inner = np.round(theta_E_inner, 10)
            kwargs = {'center_x': x, 'center_y': y,
                      'Rs_outer': Rs_angle,
                      'theta_E_inner': theta_E_inner,
                      'alpha_Rs_outer': alpha_Rs_envelope,
                      'r_trunc': r_trunc_arcsec}
            self._lenstronomy_args = [kwargs]
        return self._lenstronomy_args, None

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['CORE_COLLAPSED_HALO_BH']

    @property
    def profile_args(self):
        """

        :return:
        """
        if not hasattr(self, '_profile_args'):
            # solve for lens model parameters that conserve mass inside rs and r200
            gamma_inner = self._args['gamma_inner']
            gamma_outer = self._args['gamma_outer']
            rt_kpc = self._args['rt_kpc']
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
            sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2
            rhos, rs, r200 = self.nfw_params
            Rs_angle = rs / kpc_per_arcsec
            r_trunc_angle = rt_kpc / kpc_per_arcsec
            Rs_angle_inner = self._args['Rs_inner_kpc'] / kpc_per_arcsec

            rho0 = self._profile_envelope.alpha2rho0(1.0, Rs_angle)
            r_match_arcsec = self._args['r_match_kpc'] / kpc_per_arcsec
            r200_arcsec = r200 / kpc_per_arcsec
            a = self._profile_envelope.mass_3d(r_match_arcsec,
                                               Rs_angle,
                                               rho0,
                                               r_trunc_angle / Rs_angle)  # mass enclosed NFW inside r_match
            b = self._profile_center.mass_3d_lens(r_match_arcsec,
                                                  Rs_angle_inner,
                                                  1.0,
                                                  gamma_inner,
                                                  gamma_outer)  # mass enclosed NFW inside r_match
            c = self._profile_envelope.mass_3d(r200_arcsec,
                                               Rs_angle,
                                               rho0,
                                               r_trunc_angle / Rs_angle)  # mass enclosed NFW inside r_match
            d = self._profile_center.mass_3d_lens(r200_arcsec,
                                                  Rs_angle_inner,
                                                  1.0,
                                                  gamma_inner,
                                                  gamma_outer)  # mass enclosed NFW inside r_match
            x = np.array([self._args['m_target_R'], self._args['m_target_r200']])
            M_inverse = (1 / (a * d - b * c)) * np.array([[d, -b], [-c, a]])
            y = np.matmul(M_inverse, x) / sigma_crit_arcsec
            alpha_Rs_envelope, alpha_Rs_center = y[0], y[1]
            if alpha_Rs_center < 0 or alpha_Rs_envelope < 0:
                raise Exception('alpha_Rs_envelope and alpha_Rs_center cannot be negative')
            assert np.isfinite(alpha_Rs_envelope)
            assert np.isfinite(alpha_Rs_center)
            factor = self._lens_cosmo.point_mass_factor_z(self.z)
            theta_E_inner = factor * np.sqrt(self._args['m_target_R'])
            self._profile_args = (theta_E_inner, Rs_angle, alpha_Rs_envelope, Rs_angle, r_trunc_angle)

        return self._profile_args



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.cnfw import CNFW as CNFWLenstronomy
import numpy as np

class CoreNFWHalo(Halo):
    # we use the pseudo nfw methods to normalize profile
    _pseudo_nfw = False
    """
    The main class for an NFW field halo profile without truncation

    See the base class in Halos/halo_base.py for the required routines for any instance of a Halo class
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag,
                 conserve_m200=True):

        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        self._cnfw_lenstronomy = CNFWLenstronomy()
        mdef = 'CNFW'
        super(CoreNFWHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                            lens_cosmo_instance, args, unique_tag)

    def density_profile_3d(self, r, kwargs_lenstronomy=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        if kwargs_lenstronomy is None:
            kwargs_lenstronomy = self.lenstronomy_params[0][0]
        kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
        sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
        sigma_crit_kpc = sigma_crit_mpc * 1e-6
        factor = sigma_crit_kpc / kpc_per_arcsec
        return factor * self._cnfw_lenstronomy.density_lens(r / kpc_per_arcsec,
                                                            kwargs_lenstronomy['Rs'],
                                                            kwargs_lenstronomy['alpha_Rs'],
                                                            kwargs_lenstronomy['r_core'])

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['CNFW']

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        rhos_kpc, rs_kpc, r200_kpc = self.nfw_params
        kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
        Rs_angle = rs_kpc / kpc_per_arcsec
        r_core_angle = Rs_angle * self._args['beta']
        _, alpha_Rs_nfw = self._lens_cosmo.nfw_physical2angle_fromNFWparams(rhos_kpc * 1e9,
                                                                         rs_kpc * 1e-3,
                                                                         self.z,
                                                                            pseudo_nfw=False)
        mass_2d_core = self._cnfw_lenstronomy.mass_2d(rs_kpc, rs_kpc, 1.0, self._args['beta'] * rs_kpc)
        mass_2d_nfw = self._cnfw_lenstronomy.mass_2d(rs_kpc, rs_kpc, 1.0, 1e-4 * rs_kpc)
        theta_Rs = alpha_Rs_nfw * mass_2d_core / mass_2d_nfw

        x, y = np.round(self.x, 4), np.round(self.y, 4)
        Rs_angle = np.round(Rs_angle, 10)
        theta_Rs = np.round(theta_Rs, 10)
        r_core_angle = np.round(r_core_angle, 10)
        kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle, 'r_core': r_core_angle,
                  'center_x': x, 'center_y': y}]
        return kwargs, None

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            concentration, beta = self.c, self._args['beta']
            self._profile_args = (concentration, beta)
        return self._profile_args
